from stream import Stream
from operation_handler import OperationHandler

class ComponentStream(Stream):
  def sink(self, name=None, fields=None):
    self.outputs(name, fields)

  def outputs(self, name=None, fields=None, relation=None, scope=None):
    op = OperationHandler(self._app, self.__class__)
    op.build_multilang_operation("component_output", name, fields, relation, scope)\
      .add_operation_properties_to_info("name", "type", "columns", "relation", "scope")\
      .create_arc_info_from_stream(self)\
      .handle_operation()



# OPERATION Filter
# LANGUAGE_SYNTAX 
#   stream.filter( 
#     name = "name",                # optional
#     emits = "stream",             # optional
#     prepare = prepare_function,   # optional if no initialization needed
#     keep = keep_function
#   )
# SYNTAX_NOTES 
# A "filter" may only emit a single stream.
# 
# The **"prepare_function"** must take in a single argument (the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "keep_function".
# 
# The **"keep_function"** can be a full function or a lambda function. It must take in a single argument (the tuple), and return boolean "True" or "False". Tuples will pass through if "keep_function" returns "True".
# 
# EXAMPLE 
# # Checks for tuples whose "url" field begin with "https"
# def keep_https(tuple):
#   return tuple["url"].starts_with("https")
#  
# stream = stream.filter(name = "https_filter", keep = keep_keep_https)
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper

class Filter:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, emits, prepare, keep, parallelism):
    self._name = name if name else "filter_"+Counter.get()
    self._type = "filter"

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._keep = keep
    self._parallelism = parallelism

    Helper.check_name("filter", self._name, self._app._names)
    Helper.check_filter(self)
    Helper.check_emits("filter", self._emits, self._app._streams)

  def run_operation(self):

    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          if self._keep(tup):
            controller.emit(tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in filter: "+traceback.format_exc())
      controller.done()



import signal
import os
import sys
import json
import datetime
import base64
import socket
from tup import *

class UserEndCycleException(Exception):
  pass

class ParentDeadException(Exception):
  pass

class Controller:

  READ_SOFT_TIMEOUT = 20 # Amount of time to wait until we send a 'ping' to the parent
  DEATH_TIMEOUT = READ_SOFT_TIMEOUT * 3  # If no activity from the parent (even response to a ping) time before we kill ourself.

  def __init__(self, emits, options = {}):
    
    self.last_activity = None
    self.emits = emits
    host = options.get("host", None)
    port = options.get("port", None)
    unix_socket = options.get("unix_socket", None)
    
    if(host != None and port != None):
      sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
      sock.connect((host, int(port)))
      self.socket_to_java = sock.makefile('r+w', 1)
      self.socket_from_java = self.socket_to_java
    elif(unix_socket != None):
      sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
      sock.connect(unix_socket)
      self.socket_to_java = sock.makefile('r+w', 1)
      self.socket_from_java = self.socket_to_java
    else:
      self.socket_to_java = sys.stdout
      self.socket_from_java = sys.stdin

  def pjson(self, m):
    jd = json.dumps(m)
    self.socket_to_java.write(jd+"\n")
    self.socket_to_java.write("end\n")
    self.socket_to_java.flush()

  def emit(self, *args):
    # convert tuple to json, print to stdout
    m = {"command":"emit"}
    m["stream"] = None
    m["tuple"] = {}
    m["meta"] = {}

    if(len(args) == 1):
      m["tuple"] = args[0]
    elif(len(args) == 2):
      if isinstance(args[0], basestring):
        m["stream"] = args[0]
        m["tuple"] = args[1]
      elif isinstance(args[0], dict):
        m["tuple"] = args[0]
        m["meta"] = args[1]
    elif(len(args) == 3):
      m["stream"] = args[0]
      m["tuple"] = args[1]
      m["meta"] = args[2]

    if isinstance(m["tuple"], Tuple):
      if m["tuple"].meta != {}:
        m["meta"] = m["tuple"].meta
      m["tuple"] = m["tuple"].values

    if(m["stream"] == None):
      m["stream"] = self.emits[0]

    self.pjson(m)

  def log(self, message):
    # convert to json, print to stdout
    m = {"command":"log"}
    m["msg"] = message
    self.pjson(m)

  def ack(self, idn):
    if idn == None:
      self.log("Invalid id")
    else:
      m = {"command":"ack", "id":idn}
      self.pjson(m)

  def fail(self, message):
    if message == None:
      self.log("no error message")
    else:
      m = {"command":"fail", "msg":message}
      self.pjson(m)

  def done(self):
    m = {"command":"done"}
    self.pjson(m)

  def end_cycle(self):
    m = {"command":"end_cycle"}
    self.pjson(m)
    raise UserEndCycleException("cycle ended")

  def check_end(self):
    line = self.socket_from_java.readline()
    while line != "end\n":
      self.log("Missing end checking next line")
      line = self.socket_from_java.readline()
    return True

  def read(self):
    def timed_read(c, m):
      try:
        line = c.socket_from_java.readline()
        try:
          d = json.loads(line)
          m["msg"] = d
        except:
          c.log("Cannot be parsed to JSON")
          m["msg"] = None
        c.check_end()
        c.last_activity = datetime.datetime.now()
      except KeyboardInterrupt:
        raise
      except: # This happens on time-out (set below)
        if c.last_activity + datetime.timedelta(0,Controller.DEATH_TIMEOUT) < datetime.datetime.now():
          raise ParentDeadException("jvm appears to be dead")
        else:
          c.pjson({"ping":(datetime.datetime.utcnow()-datetime.datetime(1970,1,1)).total_seconds()})
        msg_from_java["msg"] = None

    if self.last_activity == None:
      self.last_activity = datetime.datetime.now()

    signal.signal(signal.SIGALRM, timed_read)
    signal.alarm(Controller.READ_SOFT_TIMEOUT)

    msg_from_java = {}
    timed_read(self, msg_from_java)

    if msg_from_java["msg"] != None and "pong" in msg_from_java["msg"]: # ignore pongs
      msg_from_java["msg"] = None

    return msg_from_java["msg"]

  def get_tuple(self, d):
    try:
      tup = d["tuple"]
    except:
      self.log("Missing tuple field in emitted JSON object")
      return None
    try:
      meta = d["meta"]
    except:
      self.log("Missing meta field in emitted JSON object")
      return None
    full_tuple = Tuple(d)
    return full_tuple

  def get_pidDir(self):
    d = self.read()
    cmd = d["pidDir"]
    m = {}
    m["pid"] = os.getpid()
    self.pjson(m)



# OPERATION Call Component 
# LANGUAGE_SYNTAX 
#   stream.call_component( 
#     name = "name",                                           # optional
#     component_id = "component_id_or_name",
#     additional_inputs = [other_input_stream_object_1, ...],  # blank if none
#     outputs = ["output_stream_name_1", ...],                 # optional for single output stream
#     output_format = "replace" OR "merge"                     # optional, defaults to "replace"
#   )
# SYNTAX_NOTES 
#  
#  The "component_id" MUST be given and correspond to the id listed in the output of "zillabyte components".
#  
#  The allowed output formats are "replace" and "merge". Note that only linear components, i.e. those with only "each" and "filter" operations support "merge".
#    * **"replace"** : discards the input tuple values and only emits the output values. This is the default.
#    * **"merge"** : re-emits the input tuple values along with the output values.
#  
#  To correctly stich in the component, the implicit assumptions below WILL BE USED:
#    * The "stream" that "call_component" is invoked on MUST correspond to the first listed input stream to the component.
#    * The streams specified in "additional_inputs" MUST correspond to the other listed input streams in order.
#    * Tuples emitted from the preceeding operation MUST contain the fields listed for the corresponding component input streams.
#    * The streams specified in "outputs" must correspond to the listed output streams to the component in order.
#    * The number of input and output streams specified must match the number listed for the component.
# EXAMPLE 
#  import zillabyte
#  
# dictionary = ["apple", "butterfly", "bus"]
# count = 0
#  
# # next_tuple for source
# def nt(controller):
#   global count
#   global dictionary
#   controller.emit({"word": dictionary[count % len(dictionary)]})
#   count += 1
#  
#   if (count == 5):
#     controller.end_cycle()
#  
#  
# app = zillabyte.app(name="plural_app")
#  
# stream = app.source(name="source", next_tuple=nt, end_cycle_policy="explicit")
# # We call the "pluralize_component" which pluralizes a incoming "word" and outputs it to the "pluralized_words" stream
# stream = stream.call_component(component_id="pluralize", outputs=["pluralized_words"])
#  
# stream.sink(name="plural_words", columns=[{"word":"string"}])
from counter import Counter
from helper import Helper

class InjectedComponent:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, idn, input_stream_1, additional_inputs, outputs, output_format):
    self._name = name if name else "component_"+Counter.get()
    self._id = str(idn)
    self._type = "component"
    self._consumes = [input_stream_1]

    if isinstance(additional_inputs, list):
      self._consumes += additional_inputs
    elif additional_inputs != None:
      self._consumes.append(additional_inputs)

    if outputs == None:
      outputs = ["stream_"+Counter.get()]
    elif isinstance(outputs, basestring):
      outputs = [outputs]
    self._emits = outputs

    self._output_format = output_format

    Helper.check_name("component", self._name, self._app._names)
    Helper.check_call_component(self)
    Helper.check_emits("component", self._emits, self._app._streams)

  def run_operation(self):
    print "Embedded components are only instantiated on our cluster, sorry!"



# OPERATION Join
# LANGUAGE_SYNTAX 
#   lhs_stream.join_with( 
#     name = "name",                                                   # optional
#     stream = rhs_stream_object,
#     fields = "join_field" OR ["lhs_join_field", "rhs_join_field"],
#     emits = "stream",                                                # optional
#     options = {options} 
#   )
# SYNTAX_NOTES 
# A "join" may only emit a single stream.
# 
# The following options can be used in a join:
# 
# Mandatory:
# 
#  * **"fields"** : specifies the fields to join on. The value must be a STRING or a length-2 ARRAY. If value = a STRING, the LH and RH join fields will both be set to this STRING. If value = a length-2 ARRAY, the LH join field will be set to array[0] and the RH join field will be set to array[1].
# 
# Optional:
#  * **"type"** : specifies the join type. The default is "left". Options are ["inner", "outer", "left", "right"]
#  * **"emits"** : specifies the stream to emit. A "join" may only emit a single stream.
# 
# EXAMPLE 
#  
# word_stream = employee_stream.join_with(\
#   name = "join",\
#   stream = budget_stream,\
#   fields = ["occupation", "number"],\
#   options = {"type" : "left"}\
# )
from counter import Counter
from helper import Helper
class Join:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, lhs_stream, rhs_stream, emits, fields, options):
    self._name = name if name else "join_"+Counter.get()
    self._type = "join"
    self._lhs_stream = lhs_stream._name
    self._rhs_stream = rhs_stream._name

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    if fields:
      if isinstance(fields, list) and len(fields) == 2:
        self._lhs_fields = fields[0]
        self._rhs_fields = fields[1]
      elif isinstance(fields, basestring):
        self._lhs_fields = fields
        self._rhs_fields = fields

    self._join_type = "inner"
    if options:
      if options["type"]:
        self._join_type = options["type"]

    Helper.check_name("join", self._name, self._app._names)
    Helper.check_join(self)
    Helper.check_emits("join", self._emits, self._app._streams)

  def run_operation(self):
    print "Joins can only be instantiated on our cluster, sorry!"



# OPERATION Each
# LANGUAGE_SYNTAX 
#   stream.each( 
#     name = "name",                            # optional
#     emits = ["stream_1", "stream_2", ...],    # optional for single output stream
#     output_format = "replace" OR "merge",     # optional, defaults to "replace"
#     prepare = prepare_function,               # optional if no initialization needed
#     execute = execute_function
#   )
# SYNTAX_NOTES 
# The allowed output formats are "replace" and "merge".
#   * **"replace"** : discards the input tuple values and only emits the specified values. This is the default.
#   * **"merge"** : re-emits the input tuple values along with the specified values.
# 
# The **"prepare_function"** must take in a single argument (the "controller") and should return nothing. This is where any setup is done to prepare for tuple processing in the "execute_function".
# 
# The **"execute_function"** must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the tuples are actually processed.
#  
# EXAMPLE 
# from mechanize import Browser
#  
# br = None
# def prep(controller):
#   global br
#   br = Browser()
#  
# # Open a url and emit the "url" and "title" fields
# def ee(controller, tup):
#   global br
#   url = tup["url"]
#   br.open(url)
#   controller.emit({"url":url, "title":br.title()})
# 
# # call the each
# stream = stream.each(name = "test_python", prepare = prep, execute = ee)
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper

class Each:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, emits, output_format, prepare, execute, parallelism):
    self._name = name if name else "each_"+Counter.get()
    self._type = "each"
    self._output_format = output_format

    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._execute = execute
    self._parallelism = parallelism

    Helper.check_name("each", self._name, self._app._names)
    Helper.check_each(self)
    Helper.check_emits("each", self._emits, self._app._streams)

  def run_operation(self):
    
    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" in d and d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        else:
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._execute(controller, tup)
      except KeyboardInterrupt:
        raise
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except:
        controller.fail("Exception in each: "+traceback.format_exc())
      controller.done()



# OPERATION Sink 
# LANGUAGE_SYNTAX  
#   stream.sink( 
#     name = "relation_name",
#     columns = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...]
#   )
# SYNTAX_NOTES 
# "Sink" relation "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
#  
# "Columns" must be a non-empty LIST.
#  
# Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
#  
# Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
#  
# Field types may be "string", "integer", "float", "double", "boolean", "array" or "map".
# EXAMPLE 
# stream.sink(name = "test_python_sink", columns = [{"url":"string"},{"title":"string"}])
from helper import Helper

class Sink:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, columns, relation, scope):
    self._name = name
    self._type = "sink"
    self._columns = columns

    if relation != None:
      self._relation = relation
    else:
      self._relation = name
    self._scope = scope

    Helper.check_sink(self._name, self._columns, self._app._nodes, "sink")

  def run_operation(self):
    print "Sinks cannot be run!"



# OPERATION Group By
# LANGUAGE_SYNTAX 
#   stream.group_by( 
#     name = "name",                            # optional
#     fields = ["field_1", "field_2", ...],
#     emits = ["stream_1", "stream_2", ...],    # optional for single output stream
#     begin_group = begin_group_function,
#     aggregate = aggregate_function,
#     end_group = end_group_function 
#   )
# SYNTAX_NOTES 
# The **"begin_group_function"** must take in 2 arguments (the "controller" and the "grouping tuple", which is emitted at the beginning of each group and contains the values of the fields specified in "group_by"), and should return nothing. This is where initial values for the aggregation are set.
# 
# The **"aggregate_function"** must take in 2 arguments (the "controller" and the "tuple"), and should return nothing. This is where the aggregation is performed.
# 
# The **"end_group_function"** must take in a single argument (the "controller"), and should return nothing. This is where the final aggregated value is emitted.
# EXAMPLE 
#  
# # The begin_function is where you initialize any state
# # In this case, we save the ["word", "url"] pair we wish to count
# def word_count_begin_group(controller, g_tuple):
#   global wc_word
#   global wc_url
#   global wc_count
#   wc_word = g_tuple["word"]
#   wc_url = g_tuple["url"]
#   wc_count = 0
#  
# # In this simple case, increment the counter
# def word_count_aggregate_group(controller, tup):
#   global wc_count
#   wc_count += 1
#  
# # This is run after all tuples have been received for the cycle
# # We emit the "word" , "url", and the count of the pair in tuples
# def word_count_end_group(controller):
#   global wc_word
#   global wc_url
#   global wc_count
#   controller.emit({"word" : wc_word, "url" : wc_url, "count" : wc_count}) 
#  
# # This is how the stream is actually declared
# word_stream = stream.group_by(\
#   name = "word_count",\
#   fields = ["word", "url"],\
#   begin_group = word_count_begin_group,\
#   aggregate = word_count_aggregate_group,\
#   end_group = word_count_end_group\
# )
import traceback
from counter import Counter
from controller import Controller, ParentDeadException
from helper import Helper

class GroupBy:
  def __init__(self, app):
    self._app = app
  
  def build_node(self, name, emits, fields, prepare, begin_group, aggregate, end_group, parallelism):
    self._name = name if name else "group_by_"+Counter.get()
    self._type = "group_by"

    self._group_by = fields
    if emits == None:
      emits = ["stream_"+Counter.get()]
    elif isinstance(emits, basestring):
      emits = [emits]
    self._emits = emits

    self._prepare = prepare
    self._begin_group = begin_group
    self._aggregate = aggregate
    self._end_group = end_group
    self._parallelism = parallelism

    Helper.check_name("group_by", self._name, self._app._names)
    Helper.check_group_by(self)
    Helper.check_emits("group_by", self._emits, self._app._streams)

  def run_operation(self):
    
    controller = Controller(self._emits, self._app._options)
    controller.get_pidDir()

    while True:
      try:
        d = controller.read()
        if d == None:
          continue
        elif "command" not in d:
          controller.log("Not a command "+str(d))
        elif d["command"] == "prepare":
          if self._prepare != None:
            self._prepare(controller)
        elif d["command"] == "begin_group":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._begin_group(controller, tup)
        elif d["command"] == "aggregate":
          tup = controller.get_tuple(d)
          if tup == None:
            continue
          self._aggregate(controller, tup)
        elif d["command"] == "end_group":
          self._end_group(controller)
        else:
          controller.log("Invalid command")
      except ParentDeadException, e:
        controller.log("jvm appears to have died")
        raise
      except KeyboardInterrupt:
        raise
      except:
        controller.fail("Exception in group_by: "+traceback.format_exc())
      controller.done()



# OPERATION Component Output
# LANGUAGE_SYNTAX 
#   component_stream.outputs( 
#     name = "output_stream_name",
#     fields = [{"field_1" : "type_1"}, {"field_2" : "type_2"}, ...] 
#   )
# SYNTAX_NOTES 
#   "Outputs" stream "name" must be specified as a non-empty STRING with only alphanumeric and underscore characters!
#  
#   "Fields" must be a non-empty LIST.
#  
#   Field names must be non-empty STRINGS with only alphanumeric or underscore characters.
#  
#   Field names cannot be "v[number]", "id", "confidence", "since" or "source" which are reserved Zillabyte names.
#  
#   Field types may be "string", "integer", "float", "double", "boolean", "array" or "map".
# EXAMPLE 
# stream.outputs(name = "output_stream", fields = [{"word" : "string"}])
from helper import Helper

class ComponentOutput:
  def __init__(self, app):
    self._app = app

  def build_node(self, name, columns, relation, scope):
    self._name = name
    self._type = "output"
    self._columns = columns

    if relation != None:
      self._relation = relation
    else:
      self._relation = name
    self._scope = scope

    Helper.check_sink(self._name, self._columns, self._app._nodes, "outputs")

  def run_operation(self):
    print "Component sinks cannot be run!"
